"""CGI-savvy HTTP Server.

This module builds on HTTPServer by implementing GET and POST
requests to cgi-bin scripts.

If the os.fork() function is not present (e.g. on Windows),
os.popen2() is used as a fallback, with slightly altered semantics; if
	hat function is not present either (e.g. on Macintosh), only Python
scripts are supported, and they are executed by the current process.

In all cases, the implementation is intentionally naive -- all
requests are executed sychronously.

SECURITY WARNING: DON'T USE THIS CODE UNLESS YOU ARE INSIDE A FIREWALL
-- it may execute arbitrary Python code or external programs.


***modified from python original***
doesn't really load the file from disk -- it runs modules from a short list.
makes security bit better and plays nice with the py2exe windows compilation. Also, it sends a "forever cache" Expires header to the browser for static files like the stylesheet and logo.



todo: disallow any clients that are not loopback (ie, from 127.0.0.0/8)?


todo: the base, simple, and CGI HTTPServers adhere to the HTTP and CGI specs
very well -- but are a bit inflexible. rewrite with threading, OOP, etc in mind?
"""


__version__ = "0.4"

__all__ = ["HTTPRequestHandler"]

import os
import sys
import urllib
import SimpleHTTPServer, SocketServer
import select
import meta
import util
import thread
import time

## "approved CGIs", actually modules, a la java servlet classes.
## they must expose a run() method
import search, click, cache, browse, browse_history, index, prefs
import prefs_save, addurl, delete, browse_save##, gutenberg
#import index, browse
import filters, filters_save


class HTTPRequestHandler(SocketServer.ThreadingMixIn, SimpleHTTPServer.SimpleHTTPRequestHandler):

	##approved cgis
	##todo; use gui.getattr for this... no need to load them all.
	TheList = {'search': search, 'click': click, 'cache': cache, 
	'browse': browse, 'browse_history': browse_history, 'index': index, 
	'index.html':index, 'index.htm':index, '':index, 'prefs':prefs, 'prefs_save': 
	prefs_save, 'addurl':addurl, 'delete':delete, 'browse_save':browse_save,
	'filters':filters, 'filters_save':filters_save}
	##, 'gutenberg':gutenberg}


	"""Complete HTTP server with GET, HEAD and POST commands.
		[no PUT or DELETE --Ed.]
	GET and HEAD also support running CGI scripts.

	The POST command is *only* implemented for CGI scripts.

	"""

	# Determine platform specifics
	have_fork = hasattr(os, 'fork')
	have_popen2 = hasattr(os, 'popen2')
	have_popen3 = hasattr(os, 'popen3')

	# Make rfile unbuffered -- we need to read one line and then pass
	# the rest to a subprocess, so we can't use buffered input.
	rbufsize = 0

	##post is disallowed
	def do_POST(self):
		prefs = util.prefs
		if (not prefs['allow_remote_clients']) and self.client_address[0] != '127.0.0.1':
			return

		self.send_response(501, 'Not implemented')
	
	
	def send_head(self):
		prefs = util.prefs
		if (not prefs['allow_remote_clients']) and self.client_address[0] != '127.0.0.1':
			return
	
		"""Version of send_head that support CGI scripts"""
		script = self.is_cgi()
		
		if (script != None):
			return self.run_cgi(script)
		else:
			return self.send_head_orig()
		
	## exactly the same as in SimpleHTTPServer.send_head(), but adding a couple
	## headers to tell the client to cache 'static' files. This reduces the load
	## on this server process (which is not the most robust, esp on Windows). 
	## Also makes the general GUI response snappier.
	def send_head_orig(self):
		"""Common code for GET and HEAD commands.

		This sends the response code and MIME headers.

		Return value is either a file object (which has to be copied
		to the outputfile by the caller unless the command was HEAD,
		and must be closed by the caller under all circumstances), or
		None, in which case the caller has nothing further to do.

		"""
		path = self.translate_path(self.path)
		f = None
		if os.path.isdir(path):
			for index in "index.html", "index.htm":
				index = os.path.join(path, index)
				if os.path.exists(index):
					path = index
					break
			else:
				return self.list_directory(path)
		ctype = self.guess_type(path)
		if ctype.startswith('text/'):
			mode = 'r'
		else:
			mode = 'rb'
		try:
			f = open(path, mode)
		except IOError:
			self.send_error(404, "File not found")
			return None
		self.send_response(200)
		self.send_header("Content-type", ctype)
		self.send_header("Content-Length", str(os.fstat(f.fileno())[6]))
		  
		self.send_cache_headers()
		
		self.end_headers()
		return f			

	def send_cache_headers(self):
		## tell the browser to set fantastically long cache-times for static files
		self.send_header("Expires", "Sat, 31 Oct 2037 23:23:59 GMT")		  
		self.send_header("Cache-control","max-age=86400000, public")
		## ...at which point either the image gets reloaded, Harry Houdini comes
		## back from the dead or stonehenge overflows its buffer
		## todo: will this prevent changes to dowser.css over different versions?
		## whatif the user never does a ctrl+r?
		

	def is_cgi(self):
		global TheList
		rest = self.path
		i = rest.rfind('?')
		if i >= 0:
			rest, query = rest[:i], rest[i+1:]
		else:
			query = ''
		i = rest.rfind('/')
		if i >= 0:
			script, rest = rest[:i], rest[i+1:]
		else:
			script, rest = rest, ''
		
		self.cgi_info = script, rest + "?" + query
		
		try:
			return self.TheList[rest]
		except:
			return None


	def run_cgi(self, script_module):
	
		dir, rest = self.cgi_info
		i = rest.rfind('?')
		if i >= 0:
			rest, query = rest[:i], rest[i+1:]
		else:
			query = ''
		i = rest.find('/')
		if i >= 0:
			script, rest = rest[:i], rest[i:]
		else:
			script, rest = rest, ''
		
		scriptname = rest
		
		
		#print query
		
		cgi_directories = ['/gui']

		# Reference: http://hoohoo.ncsa.uiuc.edu/cgi/env.html
		# XXX Much of the following could be prepared ahead of time!
		env = {}
		env['SERVER_SOFTWARE'] = self.version_string()
		env['SERVER_NAME'] = self.server.server_name
		env['GATEWAY_INTERFACE'] = 'CGI/1.1'
		env['SERVER_PROTOCOL'] = self.protocol_version
		env['SERVER_PORT'] = str(self.server.server_port)
		env['REQUEST_METHOD'] = self.command
		uqrest = urllib.unquote(rest)
		env['PATH_INFO'] = uqrest
		env['PATH_TRANSLATED'] = self.translate_path(uqrest)
		env['SCRIPT_NAME'] = scriptname
		if query:
			env['QUERY_STRING'] = query
		host = self.address_string()
		if host != self.client_address[0]:
			env['REMOTE_HOST'] = host
		env['REMOTE_ADDR'] = self.client_address[0]
		
		# XXX AUTH_TYPE
		# XXX REMOTE_USER
		# XXX REMOTE_IDENT
		if self.headers.typeheader is None:
			env['CONTENT_TYPE'] = self.headers.type
		else:
			env['CONTENT_TYPE'] = self.headers.typeheader
		length = self.headers.getheader('content-length')
		if length:
			env['CONTENT_LENGTH'] = length
		accept = []
		for line in self.headers.getallmatchingheaders('accept'):
			if line[:1] in "  \n\r ":
				accept.append(line.strip())
			else:
				accept = accept + line[7:].split(',')
		env['HTTP_ACCEPT'] = ','.join(accept)
		ua = self.headers.getheader('user-agent')
		if ua:
			env['HTTP_USER_AGENT'] = ua
		co = filter(None, self.headers.getheaders('cookie'))
		if co:
			env['HTTP_COOKIE'] = ', '.join(co)
		# XXX Other HTTP_* headers
		if not self.have_fork:
			# Since we're setting the env in the parent, provide empty
			# values to override previously set values
			for k in ('QUERY_STRING', 'REMOTE_HOST', 'CONTENT_LENGTH',
					  'HTTP_USER_AGENT', 'HTTP_COOKIE'):
				env.setdefault(k, "")
		
		
		##os.environ.update(env)

		self.send_response(200, "Script output follows")


		decoded_query = query.replace('+', ' ')
		
		###XXX spawn a thread rather than fork a process?
		##yes. :)
		meta.results = []
		##thread.start_new_thread(script_module.run, 
		##	(env, util.parseargs(env), self.wfile))

		script_module.run(env, util.parseargs(env), self.wfile)
		return						
		
"""
		if self.have_fork:
			# Unix -- fork as we should
			args = [script]
			if '=' not in decoded_query:
				args.append(decoded_query)
			nobody = nobody_uid()
			self.wfile.flush() # Always flush before forking
			##pid = os.fork() ##XXX FORK
			
			
		
			meta.results = []
			script_module.run(env, self.wfile)
			
			time.sleep(300)
			return						
				
			# Child
			try:
				
				try:
					os.setuid(nobody)
				except os.error:
					pass
					
				##xx right here -- instead of reading a file and executing
				os.dup2(self.rfile.fileno(), 0)
				os.dup2(self.wfile.fileno(), 1)
				
				##os.execve(scriptfile, args, os.environ)
				script_module.run(env, self.wfile)
								
				os._exit(0)
				
			except:
				self.server.handle_error(self.request, self.client_address)
				os._exit(127)



		else:

  
			
			# throw away additional data [see bug #427345]
			while select.select([self.rfile._sock], [], [], 0)[0]:
				if not self.rfile._sock.recv(1):
					break
			##shutil.copyfileobj(fo, self.wfile)
			self.rfile.close()

			#sys.stdout = self.wfile
			#sys.stdout.flush()

			
			meta.results = []
			script_module.run(env, self.wfile)
"""


nobody = None

def nobody_uid():
	"""Internal routine to get nobody's uid"""
	global nobody
	if nobody:
		return nobody
	try:
		import pwd
	except ImportError:
		return -1
	try:
		nobody = pwd.getpwnam('nobody')[2]
	except KeyError:
		nobody = 1 + max(map(lambda x: x[2], pwd.getpwall()))
	return nobody


def executable(path):
	"""Test for executable file."""
	try:
		st = os.stat(path)
	except os.error:
		return False
	return st.st_mode & 0111 != 0



