#!/usr/bin/python
"""
auto-summarizer library.

	sumer.summarize(text, [percent, [lang]])

This is naive but fairly effective. todo: needs stemming, 
multilingual support. heavy HTML can defeat this easily. It
works best on articles, not message digests, indexes, and other
pathological texts.
"""
import string, re, gui.search, msg

sum_cutoff = 2048

##todo: dedupe?
def summarize(text, percent=0.2, lang='en'):
	target_len = int(len(text) * percent)
	sents = split_sent(text)

	for i in range(len(sents)):
		## sentence, score, position, length
		sents[i] = [sents[i], 0 , i, len(sents[i])]
	
	## create a table of words and their counts
	##todo: setdefault()?
	words_raw = split_word(text)
	words_raw = filter(msg.stopword, words_raw)
	words = {}
	for w in words_raw:
		if words.has_key(w):
			words[w] = words[w] + 1
		else:
			words[w] = 1
	
	##rotate, sort, weed out, reverse sort, rotate back
	word_rank = [(v, k) for k, v in words.items()]
	word_rank.sort()
	word_rank = filter(mincount, word_rank)
	word_rank.reverse()
	word_rank = [(v, k) for k, v in word_rank]
	
	##scoring
	for w, cnt in word_rank:
		for s in sents:
			if string.find(s[0], w) > -1:
				s[1] = s[1] + cnt
	
	##rearrange, sort, etc.
	sents = [(score, pos, text, length) for (text, score, pos, length) in sents]
	sents.sort()
	sents.reverse()
	
	##trim to given percentage
	current_len = 0
	cutoff = 0
	for i in range(len(sents)):
		current_len = current_len + sents[i][3]
		if current_len >= target_len or current_len >= sum_cutoff:
			cutoff = i
			break
	if cutoff > 0:
		sents = sents[0:cutoff+1]
	
	
	sents = [(pos, score, text, length) for (score, pos, text, length) in sents]
	sents.sort()
	
	lastidx = -2
	buf = []
	for i in range(len(sents)):
		buf.append(sents[i][2])
		if sents[i][0] - lastidx == 2:
			buf.append("... ")
		if sents[i][0] - lastidx > 2:
			buf.append("... <br /><br/>")
		lastidx = sents[i][0]
	
	##keywords
	word_rank = [k for k, v in word_rank]
	word_rank = filter(msg.stopword, word_rank)
	if len(word_rank) > 8:
		word_rank=word_rank[0:8]
	
	return [" ".join(buf), word_rank]
	

"""
filter for single occurances. Summaries tend to favor long
sentences. todo: experimental!
"""
def mincount(x):
	if x[0] > 1:
		return x

"""
Splits a document into sentences, defined as ".", "?", or "!"
followed by whitespace. This is naive, as it will split on
acronyms ("Dr.", "Mrs.") and may not handle quoted text well.
"""
def split_sent(text):
	r_sent = re.compile(r"([\.\?\!]+)\s?")
	s = r_sent.sub(r"\1 ___MARK_SENT__", text)
	s = re.compile("___MARK_SENT__").split(s)
	s = map(string.strip, s)
	return s
	
def split_word(text):
	r_word = re.compile(r"[^\w]+")
	w = r_word.split(text)
	w = map(string.strip, w)
	w = filter(msg.stopword, w)
	w = map(string.lower, w)
	return w
	

def test():
	text = """
	First of all let me say that as a Democrat I'm ashamed that the people using this information for intimidation purposes share my desire to change Presidents. This amounts to nothing more than bully tactics that I might expect from other Political Parties, but not from mine. Freedom of expression and political affiliation is what we are built on and hsould never be threatened.

Having said that I don't understand this on two fronts ..

1. What possible benefit could the list could be to somebody? I mean it's not like these are potential swing votes. To me it would only give the Republicans political fodder for demonizing us democrats.

2. Why is simply posting it illegal? From the article ...
The subpoena seeks subscriber information, and contacts and billing records for the Indy Media site. It says the information is needed to investigate possible violations of the federal criminal code barring efforts to intimidate, threaten or coerce voters.
Seems to me this goes along the lines of writing a phone book. Sure it's a select phone book with more detailed information (email, hotel, etc) but it's still public information (except possibly the hotel). To me it's the people who abuse the information that would intimidate, threaten or coerce voters. 

*warning* shameless plug to get myself (and yourself) a free ipod follows:
--
Stick it to the advertisers and get a free ipod [freeipods.com]
	
	"""
	for thing in summarize(text, 0.4):
		print thing[0]
	
##test()